/*
 * main.c
 *
 *  original version:
 *  Christophe Lucas <c.lucas@ifrance.com>
 *  new release:
 *  Luca Amore (lookee) <luca@lookee.it>
 *  This program is relased under GNU/GPL term :-)
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <pwd.h>
#include <signal.h>

#include <linux/compiler.h>
#include <linux/videodev.h>

#include "config.h"
#include "bmp2jpg.h"
#include "ftpup.h"

/*
 * TODO:
 *  - open webcam.conf from /etc/sn9102-webcamd.conf 
 *     and afterwards from $CWD.
 *  - use syslog from information about daemon health.
 *  - put user identity in conf file 
 *  - find another way to save rgb24 to BMP unless SDL_SaveBMP()
 */

#define CONF_FILE "webcam.conf"

int ftp_upload = 1; 	/* (1 enabled|0 disabled) ftp upload */
int save_img = 1;		/* (1 enabked!0 disabled) image saving in current directory */

/* internal */
static unsigned int ftp_rate_counter=1;

#define argv1 	"/dev/video0"
#define USER	"clucas"

/* configuration keys readed from file */

struct config_t *config;

static void create_pid_file( void )
{
	pid_t	my_pid ;
	FILE*	fd ;

	my_pid = getpid();
	
	fd = fopen( "/var/run/sn9c102-webcamd" , "w" );
	if( NULL == fd ) {
		perror("fopen");
		exit(2);
	}
	fprintf( fd , "%d" , my_pid ); 
	flose( fd );
}

void daemon_must_die( int n )
{
	if( unlink( "/var/run/sn9c102-webcamd" ) ) {
		perror("unlink");
		exit(2);
	}
}

extern void bayer2rgb24(unsigned char *dst, unsigned char *src, long int WIDTH, long int HEIGHT) ;

int main( int argc, char **argv)
{
    int			fd1, done ;
    unsigned long int	x, y, f;
    unsigned char	*s, *d;
	int pilote_gid;
	int pilote_uid;

	struct passwd *pw_infos;

    struct v4l2_cropcap 	cropcap;
    struct v4l2_crop 		crop;
    struct v4l2_format		format;

    done = 0 ;

    x=352;
    y=288;
    f=24;

    /* reading configuration file */
    /* thanks to Christophe Lucas for his patch and help (lookee) */
    config = load_conf_from_file(CONF_FILE);

    /* local file jpg */
    strncpy(config->local_file,"./cam.jpg",sizeof(config->local_file));

    /* build name of remote_tmp_file */
    /* trunk this name with NULL is bad... I'll improve the check of length of string */
    strncpy(config->remote_tmp_file,config->remote_file,sizeof(config->remote_tmp_file));
    strncat(config->remote_tmp_file,".tmp",sizeof(config->remote_tmp_file));
    
    /* check if ftp upload is enabled */
    ftp_upload=!strncmp(config->enable_ftp,"y",sizeof(config->enable_ftp));
    
    /* end copy import keys from file */
	
	/* retrieve user rights */
	pw_infos = getpwnam( USER );
	if( !pw_infos ) {
		fprintf( stderr, "User rights retrive problem") ;
		exit(1);
	}
	pilote_uid = pw_infos->pw_uid;
	pilote_gid = pw_infos->pw_gid;

	/* ds lors, tout est OK, on peut se dtacher en tant que dmon */
	if( -1 == daemon( 1 , 1 ) ) {
		perror( "daemon" );
		exit(1);
	}

	/* acquisition des droits de groupe de l'utilisateur rel */
	if( setregid(pilote_gid,pilote_gid) ) {
		perror( "setregid" );
		exit(1);
	}

	/* acquisition des droits de l'utilisateur rel */
	if( setreuid(pilote_uid,pilote_uid) ) {
		perror( "setreuid" );
		exit(1);
	}

	/* creation du var/run/pid */
	create_pid_file();

	/* mise en place des handlers signaux */
	signal(SIGTERM, daemon_must_die);
	signal(SIGINT,  daemon_must_die);
	signal(SIGQUIT, daemon_must_die);
    
	s = (char*)malloc(x * y);
    if( NULL == s ) {
		perror("Lack of memory !! :-(");
		exit(1);
    }

    d = (char*)malloc(x * y *3);
    if( d == NULL ) {
		perror("Lack of memory !! :-(");
		exit(1);
    }

    fd1 = open(argv1, O_RDONLY);
    if( fd1 == -1 ) {
		perror("Error : file does not exist (the module of webcam is loaded?)");
		exit(1);
    }

	/* only for debug (2) */
	/* printf("main: %s\n",config->host); */

    cropcap.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if( ioctl(fd1,VIDIOC_CROPCAP,&cropcap) == -1 ) {
		fprintf(stderr, "VIDIOC_CROPCAP failed\n");
		perror("ioctl");
		exit(-1);
    }

    crop.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    crop.c.top = 0;
    crop.c.left = 0;
    crop.c.width = x*f;
    crop.c.height = y*f;
    if( ioctl(fd1,VIDIOC_S_CROP,&crop) == -1 ) {
    	fprintf(stderr, "VIDIOC_S_CROP failed\n");
		perror("ioctl");
		exit(-1);
    }

    format.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
    format.fmt.pix.width = x;
    format.fmt.pix.height = y;
    format.fmt.pix.pixelformat = V4L2_PIX_FMT_SBGGR8;
    if( ioctl(fd1,VIDIOC_S_FMT,&format) == -1 ) {
    	fprintf(stderr, "VIDIOC_S_FMT failed\n");
		perror("ioctl");
		exit(-1);
    }

    while( !done ) {
		int ret ;

		ret = read(fd1, s, x * y);
		if( ret != (x * y ) ) {
		    printf("We will have fragmented picture :-(\n" );
		}
		bayer2rgb24(d, s, x, y);
	
		/* controllo sul rate di upload o salvataggio */
		if(ftp_rate_counter==FTP_RATE)
			ftp_rate_counter=1;
			
		/* save image */
		if( (save_img|ftp_upload) && (ftp_rate_counter++ == 1 ) ) {
			/* save a surface on BMP format to file (cam.bmp) */
//			SDL_SaveBMP(screen, "cam.bmp");

			/* convert BMP to JPG format (cam.jpg) */
			/* passo argv perche imagemagick lo vuole  (?!?!?!) */
			bmp2jpg(argv);
		
			/* upload FTP dell'immagine (se richiesto) */
			if (ftp_upload)
				ftpup(config);
		}
    }
    close(fd1);
return 0;
}
